-----------------------------------------------------------------------------
-- TIMain:	Type Inference Algorithm
-- 
-- Part of `Typing Haskell in Haskell', version of September 1, 1999
-- Copyright (c) Mark P Jones and the Oregon Graduate Institute
-- of Science and Technology, 1999
-- 
-- This program is distributed as Free Software under the terms
-- in the file "License" that is included in the distribution
-- of this software, copies of which may be obtained from:
--             http://www.cse.ogi.edu/~mpj/thih/
-- 
-----------------------------------------------------------------------------

module TIMain where
import List( (\\), intersect, union )
import Id
import Kind
import Type
import Subst
import Pred
import Scheme
import Assump
import TIMonad
import Infer
import StdPrel
import Lit
import Pat
import Reduce

-----------------------------------------------------------------------------

data Expr = Var   Id
          | Lit   Literal
          | Const Assump
          | Ap    Expr Expr
          | Let   BindGroup Expr

          | Lam   Alt
          | If    Expr Expr Expr
          | Case  Expr [(Pat,Expr)]

-----------------------------------------------------------------------------
-- The following helper functions are used to construct sample programs; if we
-- change the representation of Expr above, then we need only change the
-- definitions of the following combinators to match, and do not need to
-- rewrite all the test code.

ap              = foldl1 Ap
evar v          = (Var v)
elit l          = (Lit l)
econst c        = (Const c)
elet e f        = foldr Let f (map toBg e)

toBg           :: [(Id, Maybe Scheme, [Alt])] -> BindGroup
toBg g          = ([(v, t, alts) | (v, Just t, alts) <- g ],
                   filter (not . null) [[(v,alts) | (v,Nothing,alts) <- g]])

pNil            = PCon nilCfun []
pCons x y       = PCon consCfun [x,y]

eNil            = econst nilCfun
eCons x y       = ap [ econst consCfun, x, y ]

{-
ecase           = Case
elambda         = Lam
eif             = If
-}

ecase d as      = elet [[ ("_case",
                           Nothing,
                           [([p],e) | (p,e) <- as]) ]]
                       (ap [evar "_case", d])
eif c t f       = ecase c [(PCon trueCfun [], t),(PCon falseCfun [], f)]
elambda alt     = elet [[ ("_lambda",
                           Nothing,
                           [alt]) ]]
                             (evar "_lambda")
eguarded        = foldr (\(c,t) e -> eif c t e) efail
efail           = Const ("FAIL" :>: Forall [Star] ([] :=> TGen 0))
esign e t       = elet [[ ("_val", Just t, [([],e)]) ]] (evar "_val")

eCompFrom p e c = ap [ econst mbindMfun, e, elambda ([p],c) ]
eCompGuard e c  = eif e c eNil
eCompLet bgs c  = elet bgs c
eListRet e      = eCons e eNil

-----------------------------------------------------------------------------

tiExpr :: Infer Expr Type

tiExpr as (Var i)
 = do let sc = find i as
      (ps :=> t) <- freshInst sc
      return (ps, t)

tiExpr as (Const (i:>:sc))
 = do (ps :=> t) <- freshInst sc
      return (ps, t)

tiExpr as (Lit l)
 = do (ps,t) <- tiLit l
      return (ps, t)

tiExpr as (Ap e f)
 = do (ps,te) <- tiExpr as e
      (qs,tf) <- tiExpr as f
      t       <- newTVar Star
      unify (fn tf t) te
      return (ps++qs, t)

tiExpr as (Let bg e)
 = do (ps, as') <- tiBindGroup as bg
      (qs, t)   <- tiExpr (as' ++ as) e
      return (ps ++ qs, t)

tiExpr as (Lam alt)
 = tiAlt as alt

tiExpr as (If e e1 e2)
 = do (ps,t)   <- tiExpr as e
      unify t tBool
      (ps1,t1) <- tiExpr as e1
      (ps2,t2) <- tiExpr as e2
      unify t1 t2
      return (ps++ps1++ps2, t1)

tiExpr as (Case e branches)
 = do (ps, t) <- tiExpr as e
      v       <- newTVar Star
      let tiBr (pat, f)
           = do (ps, as',t') <- tiPat pat
                unify t t'
                (qs, t'')   <- tiExpr (as'++as) f
                unify v t''
                return (ps++qs)
      pss <- mapM tiBr branches
      return (ps++concat pss, v)

-----------------------------------------------------------------------------

type Alt = ([Pat], Expr)

tiAlt :: Infer Alt Type
tiAlt as (pats, e)
       = do (ps, as', ts) <- tiPats pats
            (qs,t)  <- tiExpr (as'++as) e
            return (ps++qs, foldr fn t ts)

tiAlts :: [Assump] -> [Alt] -> Type -> TI [Pred]
tiAlts as alts t
        = do psts <- mapM (tiAlt as) alts
             mapM (unify t) (map snd psts)
             return (concat (map fst psts))

-----------------------------------------------------------------------------

type Expl = (Id, Scheme, [Alt])

tiExpl :: [Assump] -> Expl -> TI [Pred]
tiExpl as (i, sc, alts) =
    do (qs :=> t) <- freshInst sc
       ps         <- tiAlts as alts t
       s          <- getSubst
       let qs'     = apply s qs
           t'      = apply s t
           ps'     = [ p | p <- apply s ps, not (qs' ||- p) ]
           fs      = tv (apply s as)
           gs      = tv t' \\ fs
           (ds,rs) = reduce fs gs ps'
           sc'     = quantify gs (qs':=>t')
       if sc /= sc' then
           error "signature too general"
        else if not (null rs) then
           error "context too weak"
        else
           return ds

-----------------------------------------------------------------------------

type Impl   = (Id, [Alt])

restricted   :: [Impl] -> Bool
restricted bs = any simple bs
 where simple (i,alts) = any (null . fst) alts

tiImpls :: Infer [Impl] [Assump]
tiImpls as bs =
   do ts <- mapM (\_ -> newTVar Star) bs
      let is    = map fst bs
          scs   = map toScheme ts
          as'   = zipWith (:>:) is scs ++ as
          altss = map snd bs
      pss <- sequence (zipWith (tiAlts as') altss ts)
      s   <- getSubst
      let ps'     = apply s (concat pss)
          ts'     = apply s ts
          fs      = tv (apply s as)
          vss     = map tv ts'
          gs      = foldr1 union vss \\ fs
          (ds,rs) = reduce fs (foldr1 intersect vss) ps'
      if restricted bs then
          let gs'  = gs \\ tv rs
              scs' = map (quantify gs' . ([]:=>)) ts'
          in return (ds++rs, zipWith (:>:) is scs')
        else
          let scs' = map (quantify gs . (rs:=>)) ts'
          in return (ds, zipWith (:>:) is scs')

-----------------------------------------------------------------------------

type BindGroup  = ([Expl], [[Impl]])

tiBindGroup :: Infer BindGroup [Assump]
tiBindGroup as (es,iss) =
  do let as' = [ v:>:sc | (v,sc,alts) <- es ]
     (ps, as'') <- tiSeq tiImpls (as'++as) iss
     qs         <- mapM (tiExpl (as''++as'++as)) es
     return (ps++concat qs, as''++as')

tiSeq :: Infer bg [Assump] -> Infer [bg] [Assump]
tiSeq ti as []
 = return ([],[])
tiSeq ti as (bs:bss)
 = do (ps,as')  <- ti as bs
      (qs,as'') <- tiSeq ti (as'++as) bss
      return (ps++qs, as''++as')

-----------------------------------------------------------------------------
